using System;
using System.Collections;
using System.IO;

namespace InfiniTec.SystemUtilities
{
	/// <summary>
	/// This class is a replacement for the the <see cref="System.IO.MemoryStream" /> class which
	/// uses buffers from the BufferProvider. The goal is to reduce the number of objects on the
	/// Large Object Heap.
	/// </summary>
	public class SmartMemoryStream: Stream, IDisposable
	{
		/// <summary>
		/// Private variable for public property. See <see cref="MinimumRequiredSize" />
		/// </summary>
		private static int _MinimumRequiredSize = BufferProvider.LargeObjectTreshold;

		/// <summary>
		/// Gets or sets the minum size that a buffer must have to be recycled to the BufferManager class.
		/// Defaultvalue is <see cref="BufferProvider.LargeObjectTreshold" />.
		/// </summary>
		public static int MinimumRequiredSize { get { return _MinimumRequiredSize; } set { _MinimumRequiredSize = value; } }
		
		/// <summary>
		/// Stores the value for the <see cref="Length"/> of the stream.
		/// </summary>
		private long _Length;
		/// <summary>
		/// Stores the value for the current <see cref="Position"/> within the stream.
		/// </summary>
		private long _Position;
		/// <summary>
		/// Stores the valud for the current <see cref="Capacity"/> of the stream.
		/// </summary>
		private long _Capacity;
		/// <summary>
		/// Contains the actual buffer of the stream.
		/// </summary>
		private byte[] _Buffer;

		/// <summary>
		/// Returns the capacity of the stream in bytes.
		/// </summary>
		public long Capacity { get { return _Capacity; } }

		/// <summary>
		/// Returns, whether reading is possible on this stream. Always true for this claas.
		/// </summary>
		public override bool CanRead { get { return true; } }

		/// <summary>
		/// Returns, whether seeking is possible on this stream. Always true for this claas.
		/// </summary>
		public override bool CanSeek { get { return true; } }

		/// <summary>
		/// Returns, whether writing is possible on this stream. Always true for this claas.
		/// </summary>
		public override bool CanWrite { get { return true; } }

		/// <summary>
		/// Returns the length of this stream in bytes. See <see cref="Stream.Length"/>
		/// </summary>
		public override long Length { get { return _Length; } }

		/// <summary>
		/// Gets or sets the actual position within the stream. See <see cref="Stream.Position"/>
		/// </summary>
		public override long Position
		{
			get { return _Position; } 
			set
			{
				if ((value < 0) || (value > Length)) throw new ArgumentException("Position must be between 0 and Length");
				_Position = value;
			}
		}

		/// <summary>
		/// In this implementation, this method does nothing.
		/// </summary>
		public override void Flush()
		{

		}

		/// <summary>
		/// Writes the content of the buffer into the buffer of the stream.
		/// </summary>
		/// <remarks>
		/// <p>
		///	If the capacity of the current <see cref="_Buffer" /> is exceeded, a new buffer
		///	is aquired from the <see cref="BufferProvider"/>. The new buffer has at least a
		///	size of (<see cref="Capacity"/> + count)*2. The obsolete old buffer is recycled by
		///	the BufferProvider if its capacity exceeds the <see cref="MinimumRequiredSize" />
		///	</p>
		///	<p>
		///	If you know how big the stream will grow in advance, it's a good idea to call
		///	<see cref="SetLength" /> before writing to the stream. Otherwise, a great deal of
		///	memory might be wasted because the size of the underlying buffer must be increased 
		///	several times.
		///	</p>
		///	</remarks>
		/// <param name="buffer">The buffer with the data to be copied to this stream.</param>
		/// <param name="offset">The offset in the source buffer</param>
		/// <param name="count">The number of bytes to read.</param>
		public override void Write(byte[] buffer, int offset, int count)
		{
			if (buffer == null) throw new ArgumentNullException("buffer");
			if (offset < 0) throw new ArgumentOutOfRangeException("offset", "offset must be greater 0");
			if (count < 0) throw new ArgumentOutOfRangeException("count", "count must be greater 0");

			if (offset + count > buffer.Length) throw new ArgumentOutOfRangeException("count and offset must be within the bounds of the buffer");

			EnsureSize(Position + count);

			Buffer.BlockCopy(buffer, offset, _Buffer, (int) Position, count);
			if (Position + count > Length) _Length = Position + count;
			Position += count;
		}

		/// <summary>
		/// Sets the length of the buffer to the specified value.
		/// </summary>
		/// <remarks>
		/// <p>If the new length is smaller than the current length, the stream is
		/// effectivly truncated.</p>
		/// <p>If the new length exceeds the capacity of the stream, it will be 
		/// adjusted to accomodate for the new length.</p>
		/// </remarks>
		/// <param name="value">The new length of the stream.</param>
		public override void SetLength(long value)
		{

			if (value > Capacity) SetCapacity(value);

			_Length = value;
		}

		/// <summary>
		/// Sets the position within to the specified value. 
		/// </summary>
		/// <param name="offset">The new position relative to the origin.</param>
		/// <param name="origin">One of the three <see cref="SeekOrigin"/> values.</param>
		/// <returns></returns>
		public override long Seek(long offset, SeekOrigin origin)
		{
			long newPosition;

			newPosition = 0;
			switch (origin)
			{
				case SeekOrigin.Begin: newPosition = offset; break;
				case SeekOrigin.Current: newPosition = Position + offset; break;
				case SeekOrigin.End: newPosition = Length + offset; break;
			}

			Position = newPosition;

			return newPosition;
		}

		/// <summary>
		/// Reads the number of bytes specified by count and writes them into the buffer
		/// at the specified offset. For more Information and examples see <see cref="Stream.Read" />
		/// </summary>
		/// <param name="buffer">The byte buffer to copy the content to</param>
		/// <param name="offset">The offset on the buffer</param>
		/// <param name="count">The number of bytes to write</param>
		/// <returns>The number of bytes read from the stream</returns>
		public override int Read(byte[] buffer, int offset, int count)
		{
			if (buffer == null) throw new ArgumentNullException("buffer");
			if (offset < 0) throw new ArgumentOutOfRangeException("offset", "offset must be greater 0");
			if (count < 0) throw new ArgumentOutOfRangeException("count", "count must be greater 0");

			if (offset + count > buffer.Length) throw new ArgumentOutOfRangeException("count and offset must be within the bounds of the buffer");

			if (Position + count > Length) count = (int) (Length - Position);

			Buffer.BlockCopy(_Buffer, (int) Position, buffer, offset, count);

			return count;
		}


		/// <summary>
		/// The method ensures the the underlying <see cref="_Buffer"/> has at least the
		/// size specified by the size parameter.
		/// </summary>
		/// <remarks>
		/// If the capacity of the buffer is to small, the capacity will be increased.
		/// As long as the requested size is below 2 megabytes, this method will ensure that
		/// the buffer has twice the requested size. Above this limit, a factor of 1.5 is used.
		/// </remarks>
		/// <param name="size">The minimum capacity of the buffer.</param>
		protected virtual void EnsureSize(long size)
		{
			if (size < Capacity) return;

			if (size > 1048576*2) size = Convert.ToInt64(size*1.5);
			else size = size * 2;
			SetCapacity(size);
		}

		/// <summary>
		/// Sets the capacity of the stream to the requested size.
		/// </summary>
		/// <remarks>
		/// <p>
		///	If the required size exceeds the current capacity of the buffer, a new buffer is 
		///	aquired from the <see cref="BufferProvider"/>. The content of the current buffer
		///	is copied to the new buffer.
		///	</p>
		///	<p>If the size of the old buffer exceeds the <see cref="MinimumRequiredSize"/>,
		///	it will be recycled by the BufferManager.
		///	</p>
		/// </remarks>
		/// <param name="size">The requested size</param>
		protected virtual void SetCapacity(long size)
		{
			byte[] newBuffer;

			newBuffer = BufferProvider.GetBuffer(size);

			if (Length > 0) 
			{
				Buffer.BlockCopy(_Buffer, 0, newBuffer, 0, (int) Position);
			}
			if (Capacity >= MinimumRequiredSize) BufferProvider.RecycleBuffer(_Buffer);

			_Buffer = newBuffer;
			_Capacity = newBuffer.Length;
			
		}
		/// <summary>
		/// Initializes a new instance of this class with the initial capacity specified by
		/// initialCapacity.
		/// </summary>
		/// <param name="initialCapacity">The starting capacity of the Stream.</param>
		public SmartMemoryStream(int initialCapacity): base()
		{
			SetLength(initialCapacity);
		}

		/// <summary>
		/// Initializes a new instance of this class with a starting capacity of 0.
		/// </summary>
		public SmartMemoryStream(): this(0) {}

		/// <summary>
		/// Closes the Stream and recycles the internal buffer with the BufferProvider, if it
		/// exceed the <see cref="MinimumRequiredSize"/>
		/// </summary>
		public override void Close()
		{
			if (_Buffer.Length > MinimumRequiredSize)
			{
				BufferProvider.RecycleBuffer(_Buffer);
			}
			_Buffer = null;


			base.Close ();
		}

		#region IDisposable Members

		/// <summary>
		/// Calls the <see cref="Close"/> method.
		/// </summary>
		public void Dispose()
		{
			Close();
		}

		#endregion
	}
}
